from distutils.core import setup, Extension

spam = Extension('spam',
                 sources = ['spam.c'])

setup (name = 'PackageCSystem',
       version = '1.0',
       description = 'This is a package to invoke some c code',
       ext_modules = [spam])




# -*- coding: utf-8 -*-

"""enables to import modules through the web, by adding urls to the python path.
"""
#authors: Jure Vrscaj <jure@codeshift.net> (copyright (c) 2006-10),
#         송영진(MyEvan) <myevan_net@naver.com> (copyright (c) 2010), 
#         Alex Bodnaru <alexbodn@012.net.il> (copyright (c) 2010)

# This is free software; you may use, copy, modify and/or distribute this work
# under the terms of the MIT License.


# import these to avoid trying to import them here
import stringprep, array, unicodedata, _strptime

import sys, os, re
import imp, marshal
from tempfile import mkdtemp
from StringIO import StringIO
from gzip import GzipFile
from datetime import datetime, timedelta
from time import gmtime
from sqlite3 import connect, PARSE_DECLTYPES
from base64 import encodestring
from urlparse import urlsplit, urlunsplit
from httplib import HTTPSConnection
from urllib2 import HTTPDefaultErrorHandler, HTTPSHandler, HTTPError, \
    HTTPRedirectHandler, build_opener, urlopen, Request


DOT_PYO = ".pyo"
DOT_PYC = ".pyc"
DOT_PY = ".py"
PY_FILES = (DOT_PYO, DOT_PYC, DOT_PY)

DYN_SUFIX = [ext for ext, mode, type in imp.get_suffixes() \
             if type == 3 and ext.startswith('.')][0]


settings = sys.__dict__.setdefault('urlimport_settings',
                                   dict(
                                        py_version_string='$PYTHON_VERSION',
                                        ssl_cert='', 
                                        ssl_key='', 
                                        debug=0
                                   )
)

def config(**kwargs):
    """config(key=value) - Set key to value.
       config()          - Display settings.
    """
    settings.update(kwargs)
    for k,v in settings.iteritems():
        debug(" "+str(k)+"="+repr(v), lvl=1 )

def reset():
    """remove the cache contents"""
    cache_dir = settings.get('cache_dir', '')
    if cache_dir:
        for base, dirs, names in os.walk(cache_dir):
            for name in names:
                os.remove(os.path.join(base, name))
        for base, dirs, names in os.walk(cache_dir):
            for name in dirs:
                os.rmdir(os.path.join(base, name))
        os.rmdir(cache_dir)

def debug(s, pf='| |', lvl=1):
    if lvl <= settings.get('debug', 0):
        print ("%s %s" % (pf, s))


# use this to detect unmodified resources
# taken from http://diveintopython.org/http_web_services/etags.html
class DefaultErrorHandler(HTTPDefaultErrorHandler, HTTPRedirectHandler):

    def http_error_301(self, req, fp, code, msg, headers):
        response = \
            HTTPRedirectHandler.http_error_301(self, req, fp, code, msg, headers)
        response.status = code
        return response

    def http_error_304(self, req, fp, code, msg, headers):
        response = HTTPError(req.get_full_url(), code, msg, headers, fp)
        response.status = code
        return response

# use this to open https urls with client certificates.
# taken from http://www.osmonov.com/2009/04/client-certificates-with-urllib2.html
class HTTPSClientAuthHandler(HTTPSHandler, DefaultErrorHandler):

    def __init__(self, key, cert):
        HTTPSHandler.__init__(self)
        self.key = key
        self.cert = cert

    def https_open(self, req):
        # Rather than pass in a reference to a connection class, we pass in
        # a reference to a function which, for all intents and purposes,
        # will behave as a constructor
        return self.do_open(self.getConnection, req)

    def getConnection(self, host, timeout=300):
        return HTTPSConnection(host, key_file=self.key, cert_file=self.cert)


class UrlFinder:

    re_url_ok = re.compile(r'^http://|^ftp://|^https://')
    date_format = '%a, %d %b %Y %H:%M:%S %Z'

    def __init__(self, path):
        if self.re_url_ok.match(path):
            debug("%s: accepting '%s'." % (self.__class__.__name__, path), lvl=2)
            py_version_string = settings.get('py_version_string', None)
            if py_version_string is not None:
                path = path.replace(py_version_string, '%d.%d' % sys.version_info[:2])
            self.path = path.rstrip('/') + '/'
            debug("%s: will check for '%s'." % (self.__class__.__name__, self.path), lvl=4)
            # cache
            self.cache_dir = settings.setdefault('cache_dir', mkdtemp())
            if not os.path.exists(self.cache_dir):
                os.makedirs(self.cache_dir)
            self.cacheIndexDB = os.path.join(self.cache_dir, 'urlimport-cache.sqlite')
            indexConn = connect(self.cacheIndexDB, detect_types=PARSE_DECLTYPES)
            indexCursor = indexConn.cursor()
            table_exist = """select 1 
                             from sqlite_master 
                             where type = 'table' and name = 'module_access'
                          """
            table_create = """create table module_access(
                                module varchar(1024) 
                                    primary key on conflict replace,
                                findpath varchar(1024),
                                url varchar(1024),
                                __path varchar(1024),
                                tail varchar(12),
                                expires timestamp,
                                last_modified varchar(30),
                                etag varchar(30)
                              )
                           """
            if indexCursor.execute(table_exist).fetchone() is None:
                indexCursor.execute(table_create)
            indexConn.close()
        else:
            debug("%s: rejecting non-matching path item: '%s'" % (self.__class__.__name__, path), lvl=5)
            raise ImportError

    def find_module(self, fullname, mpath=None):
        """try to locate the remote module, do this:
         1) check in cache
         2) try to get fullname.py[oc]? from http://self.path/
         3) try to get fullname.DYN_SUFIX from http://self.path/
         4) try to get __init__.py[oc]? from http://self.path/fullname/
         5) update cache if should change
         6) load the module
        """

        debug("find_module %s" % fullname, lvl=2)
        head = self.path + fullname.split('.')[-1]
        file_types = [(ext, None) for ext in PY_FILES] + \
                     [(DYN_SUFIX, None)] + \
                     [('/__init__%s' % ext, head + '/') for ext in PY_FILES]

        cache_tested = False
        for tail, __path in file_types:
            url = head + tail

            use_cache = False
            if not cache_tested: # first time only
                cache_tested = True
                try:
                    cachedFile, fileData, url, __path, tail = \
                        self.read_cache(fullname)
                    debug('got from cache', lvl=1)
                    use_cache = True
                except Exception, e:
                    debug('error reading module %s from cache: %s' % (fullname, e), lvl=4)
                    pass
            try:
                if not use_cache:
                    debug("trying to get '%s'." % url, lvl=1)
                    fileData, cachedFile = \
                        self.caching_get_file(url, fullname, tail, __path)
                if tail.startswith('/__init__'):
                    cachedFile = None
                if url.endswith(DOT_PYO) or url.endswith(DOT_PYC):
                    debug("find_compiled_module: got '%s'." % url, lvl=1)
                    return UrlCompiledLoader(url, __path, cachedFile, fileData)
                elif url.endswith(DOT_PY):
                    debug("find_source_module: got '%s'." % url, lvl=1)
                    return UrlSourceLoader(url, __path, cachedFile, fileData)
                else:
                    debug("find_dynamic_module: got '%s'." % url, lvl=1)
                    return UrlDynamicLoader(url, cachedFile)
            except Exception, e:
                debug("find_module: failed to get '%s'. (%s)" % (url, e), lvl=3)

        debug("not_found_module %s" % fullname, lvl=2)

    def user_agent(self, url, host=None):
        if host is None:
            proto, user, passwd, host, port, path, query, fragment = \
                self.split_url(url)
        user_agent = settings.get('user_agent.%s' % url, 
                                  settings.get('user_agent.%s' % host, 
                                               settings.get('user_agent', None)))
        if user_agent is not None:
            py_version_string = settings.get('py_version_string', None)
            if py_version_string is not None:
                user_agent = user_agent.replace(py_version_string, '%d.%d' % sys.version_info[:2])
        return user_agent

    def no_cache(self, url, host=None):
        if host is None:
            proto, user, passwd, host, port, path, query, fragment = \
                self.split_url(url)
        return settings.get('no_cache.%s' % url, 
                            settings.get('no_cache.%s' % host, 
                                         settings.get('no_cache', False)))

    def cached_file(self, module, tail):
        return os.path.join(self.cache_dir, module.replace('.', os.path.sep)) + \
               tail

    def read_cache(self, module):
        indexConn = connect(self.cacheIndexDB, detect_types=PARSE_DECLTYPES)
        indexCursor = indexConn.cursor()
        select = """select last_modified, etag, url, __path, tail, expires 
                    from module_access 
                    where module = ? and findpath = ?
                 """
        delete = """delete module_access 
                    where module = ?
                 """
        modincache = indexCursor.execute(select, (module, self.path)).fetchone()
        if modincache is not None:
            last_modified, etag, url, __path, tail, expires = modincache
            cachedFile = fileData = None
            try:
                fileData, cachedFile = \
                    self.caching_get_file(url, module, tail, __path, 
                                          last_modified, etag, expires)
                indexConn.close()
                return cachedFile, fileData, url, __path, tail
            except:
                pass
            indexCursor.execute(delete, (module,))
            indexConn.commit()
            cachedFile = self.cached_file(module, tail)
            os.remove(cachedFile)
            os.removedirs(os.path.dirname(cachedFile))
        indexConn.close()
        raise Exception('module "%s" not in cache', module)

    def caching_get_file(self, url, module, tail, __path, \
                         cached_last_modified=None, cached_etag=None, \
                         expires=None):
        """get the file data using cache.
        """
        keep_cached = not_expired = \
            expires is not None and expires > datetime(*gmtime()[:6])
        if not keep_cached:
            fileData, expires, last_modified, keep_cached, etag, url = \
                self.get_file_data(url, cached_last_modified, cached_etag)
        cachedFile = self.cached_file(module, tail)
        if keep_cached: # assumes file is cached
            debug('reading data from cache %s:%s' % (module, cachedFile), lvl=2)
            fileData = open(cachedFile, "rb").read()
        if tail.endswith(DOT_PYC) or tail.endswith(DOT_PYO):
            if fileData[:4] != imp.get_magic():
                raise Exception("%s: incompatible python version" % url)
        if not_expired:
            return fileData, cachedFile

        proto, user, passwd, host, port, path, query, fragment = \
            self.split_url(url)
        no_cache = self.no_cache(url, host)
        if not no_cache or tail.endswith(DYN_SUFIX):
            path = os.path.dirname(cachedFile)
            if not os.path.exists(path):
                os.makedirs(path)
            if fileData is not None:
                open(cachedFile, "wb").write(fileData)
        else:
            cachedFile = None

        if not no_cache or keep_cached:
            indexConn = connect(self.cacheIndexDB, detect_types=PARSE_DECLTYPES)
            indexCursor = indexConn.cursor()
            select = """select 1
                        from module_access 
                        where module = ? and findpath = ?
                     """
            update = """update module_access
                        set 
                        url = ?,
                        __path = ?,
                        tail = ?,
                        expires = ?,
                        last_modified = ?,
                        etag = ?
                        where module = ? and findpath = ?
                     """
            update_cached = """update module_access
                               set 
                               url = ?,
                               expires = ?
                               where module = ? and findpath = ?
                            """
            insert = """insert into module_access
                        (url, __path, tail, expires,
                         last_modified, etag, module, findpath)
                        values(?, ?, ?, ?, ?, ?, ?, ?)
                     """
            if expires is None:
                cache_time = settings.get('cache_time.%s' % self.path, 
                                          settings.get('cache_time.%s' % host,
                                                       settings.get('cache_time', 
                                                                    None)))
                if cache_time is not None:
                    expires = datetime(*gmtime()[:6]) + cache_time
            cmd = fields = None
            if keep_cached:
                debug('updating cache expires to %s for: %s' % (expires, module), pf='|>>>>>|', lvl=2)
                cmd = update_cached
                fields = (url, expires, module, self.path)
            else:
                debug('saving %s in cache at: %s' % (module, cachedFile), pf='|>>>>>|', lvl=2)
                modincache = \
                    indexCursor.execute(select, (module, self.path)).fetchone()
                cmd = update if modincache is not None else insert
                fields = (url, __path, tail, expires, last_modified, etag, 
                          module, self.path)
            indexCursor.execute(cmd, fields)
            indexConn.commit()
            indexConn.close()

        return fileData, cachedFile

    def get_file_data(self, url, cached_last_modified, cached_etag):
        """Download the file data from given url.
        """
        proto, user, passwd, host, port, path, query, fragment = self.split_url(url)
        hostport = ':'.join([host, port]) if port else host
        url = urlunsplit((proto, hostport, path, query, fragment))
        key = settings.get('ssl_key.%s' % self.path, 
                           settings.get('ssl_key.%s' % host,
                                        settings.get('ssl_key', '')))
        cert = settings.get('ssl_cert.%s' % self.path, 
                            settings.get('ssl_cert.%s' % host,
                                         settings.get('ssl_cert', '')))
        opener = settings.get('opener.%s' % self.path, 
                              settings.get('opener.%s' % host,
                                           settings.get('opener', None)))
        request = Request(url)
        debug(cert + ',' + key, pf='|>>>>|', lvl=3)
        if opener is None:
            if proto == 'https' and cert:
                # handle http over ssl with client certificate
                opener = build_opener(HTTPSClientAuthHandler(key, cert))
            else:
                opener = build_opener(DefaultErrorHandler)
        creds = user + ':' + passwd if user and passwd else user
        if creds:
            base64string = encodestring('%s:%s' % (user, passwd))[:-1]
            authheader =  "Basic %s" % base64string
            request.add_header("Authorization", authheader)
        request.add_header('Accept-encoding', 'gzip')
        if cached_etag is not None:
            request.add_header('If-None-Match', cached_etag)
        if cached_last_modified is not None:
            request.add_header('If-Modified-Since', cached_last_modified)
        user_agent = self.user_agent(url, host)
        if user_agent is not None:
            debug("sending user agent:'%s'" % user_agent, pf='****', lvl=4)
            request.add_header('User-agent', user_agent)
        response = opener.open(request)

        if hasattr(response, 'status'):
            debug("got status %d" % response.status, pf='-----------')
            if response.status == 301: # permanent redirection
                if hasattr(response, 'url') and response.url is not None:
                    url = response.url
                    debug('redirected to: %s' % url, lvl=4)
        debug("got code %d" % response.code, pf='****', lvl=4)
        keep_cached = response.code == 304

        expires = response.info().get('Expires', None)
        etag = response.info().get('ETag', None)
        fileData = None

        if not keep_cached:
            fileData = response.read()
            debug('+++++++++ received data length %d: ' % len(fileData), lvl=4)
            if response.info().get('Content-Encoding', None) == 'gzip':
                compressedstream = StringIO(fileData)
                gzipper = GzipFile(fileobj=compressedstream)
                fileData = gzipper.read()
                debug('+++++++++ uncompressed data length %d: ' % len(fileData), lvl=4)
            default_last_modified = datetime(*gmtime()[:6]).strftime(self.date_format)
            last_modified = response.info().get('Last-Modified', default_last_modified)
        else:
            fileData = None
            last_modified = cached_last_modified

        return fileData, \
               self.parse_date(expires), \
               last_modified, \
               keep_cached, \
               etag, \
               url

    def parse_date(self, strdate):
        if strdate is None:
            return None
        try:
            return datetime.strptime(strdate, self.date_format)
        except:
            return None

    def split_url(self, url):
        scheme, netloc, path, query, fragment = urlsplit(url)
        user = passwd = port = ''
        if '@' in netloc:
            auth_part, host_part = netloc.split('@')
            netloc = host_part
            if ':' in auth_part:
                user, passwd = auth_part.split(':')
            else:
                user = auth_part
        if ':' in netloc:
            netloc, port = netloc.split(':')
        return scheme, user, passwd, netloc, port, path, query, fragment


class UrlDynamicLoader:
    def __init__(self, url, cachedFile):
        self.url = url
        self.cachedFile = cachedFile

    def load_module(self, fullname):
        """add the new module to sys.modules,
        execute its source and return it
        """
        #TODO: fails reloading
        mod = imp.load_dynamic(fullname, self.cachedFile)
#        mod.__file__ = self.url
#        mod.__loader__ = self
        return mod

class UrlSourceLoader:
    def __init__(self, url, path, cachedFile, data):
        self.url = url
        self.path = path
        self.cachedFile = cachedFile
        self.source = data.replace("\r\n", "\n").replace('\r', '\n')
        self._files = {}

    def load_module(self, fullname):
        """add the new module to sys.modules,
        execute its source and return it
        """
        if self.cachedFile is not None:
            debug('loading source module %s from %s' % (fullname, self.cachedFile), lvl=2)
            mod = imp.load_source(fullname, self.cachedFile)
            debug("after loading", lvl=4)
        else:
            mod = sys.modules.setdefault(fullname, imp.new_module(fullname))
            debug("load_module: executing %s's source..." % fullname, lvl=2)
            for line in self.source.split('\n'):
                debug(line, pf='|>|', lvl=4)

            exec self.source in mod.__dict__
            debug('source executed', lvl=4)

        mod.__file__ = self.url
        mod.__loader__ = self
        if self.path:
            mod.__path__ = [self.path]

        return mod

class UrlCompiledLoader:
    def __init__(self, url, path, cachedFile, data):
        self.url = url
        self.path = path
        self.cachedFile = cachedFile
        self.code = marshal.loads(data[8:])
        self._files = {}

    def load_module(self, fullname):
        """add the new module to sys.modules,
        execute it and return it
        """
        if self.cachedFile is not None:
            debug('loading compiled module %s from %s' % (fullname, self.cachedFile), lvl=-2)
            try:
                mod = imp.load_compiled(fullname, self.cachedFile)
            except Exception, e:
                debug('error %s loading %s' % (e, self.cachedFile), lvl=-2)
                raise
        else:
            mod = sys.modules.setdefault(fullname, imp.new_module(fullname))
            debug("load_module: executing %s's code..." % fullname, lvl=2)
            debug(repr(self.code), pf='|>|', lvl=4)

            exec self.code in mod.__dict__

            debug(repr(mod), pf='|>|', lvl=4)

        mod.__file__ = self.url
        mod.__loader__ = self
        if self.path:
            mod.__path__ = [self.path]

        return mod




#!/usr/bin/python -u
#  16-09-04
# v1.0.0

# auth_example.py
# A simple script manually demonstrating basic authentication.

# Copyright Michael Foord
# Free to use, modify and relicense.
# No warranty express or implied for the accuracy, fitness to purpose or otherwise for this code....
# Use at your own risk !!!

# E-mail or michael AT foord DOT me DOT uk
# Maintained at www.voidspace.org.uk/atlantibots/pythonutils.html

"""
There is a system for requiring a username/password before a client can visit a webpage. This is called authentication and is implemented by the server - it actually allows for a whole set of pages (called a realm) to be protected by authentication. This scheme (or schemes) are actually defined by the HTTP spec, and so whilst python supports authentication it doesn't document it very well. The HTTP documentation is in the form of RFCs (http://www.faqs.org/rfcs/rfc2617.html for basic and digest authentication) which are technical documents and so not the most readable !!

When I searched the web for details on authentication with python I found lots of people asking questions, but a lack of clear answers. This document and example code shows how to manually do basic authentication with python. It is an example for performing a simple operation rather than a technical document. 

I am doing it manually rather than using an auth handler because my script is a CGI which runs once for each page access. I have to store the username/passwords between each access. A 'manual' explanation also shows more clearly what is happening.

I've seen references to three authentication schemes, BASIC, NTLM and DIGEST. It's possible there are more - but BASIC authentication is overwhelmingly the most common. This tutorial/example only covers BASIC authentication although some of the details may be applicable to the other schemes.

--

In all these examples we will be using the python standard library urllib2 to fetch web pages.

A client is any program that makes requests over the internet. It could be a browser - or it could be a python program. When a client requests a web page it sends a request to the server. That request consists of headers with certain information about the request. Here we are calling these headers 'http request headers'. If the request fails to reach a server (the server name doesn't exist or there is no internet connection for example) then the request will just fail. If the request is made by python then an exception will be raised. This exception will have a 'reason' attribute that is a tuple describing the error.

The next example shows us creating a urllib2 request object, adding a fake 'User-Agent' request header and making a request. The resulting error shows what happens if you try to fetch a webpage without an internet connection ! The User-Agent request header tells the server what program is asking for the web page - some sites (e.g. google) won't allow requests from anything other than a browser... so we might have to pretend to be a browser. (Which is generally considered bad client behaviour ? so I might get my knuckles rapped for including it here).

>>> import urllib2
>>> req = urllib2.Request('http://www.google.co.uk')
>>> req.add_header('User-Agent', 'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT)')
>>> try:
	handle = urllib2.urlopen(req)
except IOError, e:
	if hasattr(e, 'reason'):
		print 'Reason : '
		print e.reason
else:
	print handle.read()                  # if we had a connection this would print the page

		
Reason : 
(7, 'getaddrinfo failed')
>>> 

The actual exception is an URLError which is a subclass of IOError - the one tested for above in the try-except block. If you did a dir(e) in the above example then you would see all the attributes of the exception object.

If however the request reaches a server then the server will send a response back. Whether or not the request succeeds the response will still contain headers from the server (or CGI script!!). These we call here 'http response headers'. If there is a problem then this response will include an error code that describes the problem. You will already be familiar with some of these codes - 404 : Page not found, 500 : Internal Server Error etc. If this happens an exception will still be raised by urllib2, but instead of a 'reason' attribute it will have a 'code' attribute. The code attribute is an integer that corresponds to the http error code. (see http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html for a *full* list of codes).

If a page requires authentication then the error code is 401. Included in the response headers is a 'WWW-authenticate' header that tells you what authentication scheme the server is using for this page *and* also something called a realm. It is rarely just a single page that is protected by authentication but a section - a 'realm' of a website. The name of the realm is included in this header line. If the client *already knows* the username/password for this realm then it can encode them into the request headers and try again. If the username/password combination are correct, then the request will succeed as normal. If the client doesn't know the username/password it should ask the user. This means that if you enter a  protected 'realm' the client effectively has to request each page twice. The first time it will get an error code and be told what realm it is attempting to access - the client can then get the right username/password for that realm (on that server) and repeat the request. 

HTTP is a 'stateless' protocol. This means that a server using basic authentication won't 'remember' you are logged in and will need to be sent the right header for every protected page you attempt to access.

Suppose we attempt to fetch a webpage protected by basic authentication.

>>> theurl = 'http://www.someserver.com/somepath/someprotectedpage.html'
>>> try:
	handle = urllib2.urlopen(theurl)
except IOError, e:
	if hasattr(e, 'code'):
		if e.code != 401:
			print 'We got another error'
			print e.code
		else:
			print e.headers
			print e.headers['www-authenticate']
		# print e.headers.get('www-authenticate', '') might be a safer way of doing this

Note the following things. We accessed the page directly from the url instead of creating a request object. If the exception has a 'code' attribute it also has an attribute called 'headers'. This is a dictionary like object with all the headers in - but you can also print it to display all the headers.  See the last line that displays the 'www-authenticate' header line which ought to be present whenever you get a 401 error.

Output from above example :

WWW-Authenticate: Basic realm="cPanel"
Connection: close
Set-Cookie: cprelogin=no; path=/
Server: cpsrvd/9.4.2

Content-type: text/html

Basic realm="cPanel"

You can see the  authentication scheme and the 'realm' part of the 'www-authenticate' header. Assuming you know the username and password you can then navigate around that website - whenever you get a 401 error with *the same realm* you can just encode the username/password into your request headers and your request should succeed.

Lets assume you need to access pages which are all in the same realm. Assuming you have got the username and password from the user, you save the name of the realm from the first access. Then whenever you get a 401 error in the same realm (from the same server !) you know the username/password to use. So the only detail left, is knowing how to encode the username/password into the request header. This is done by encoding it as a base 64 string. It doesn't actually look like clear text - but it is only the most vaguest of 'encryption'. This means basic authentication is just that - basic. Anyone sniffing your traffic who sees an authentication request header will be able to extract your username and password from it. Many websites like yahoo or ebay, use javascript hashing/encryption and other tricks to authenticate a login. This is much harder to detect and mimic from python ! You may need to use a proxy client server and see what information your browser is actually sending to the website (See http://groups.google.co.uk/groups?hl=en&lr=&ie=UTF-8&threadm=6f402501.0409100632.47e403f3%40posting.google.com&rnum=2 for suggestions of several proxy servers that can do this). 

There is a very simple recipe on the Activestate  Python Cookbook (It's actually in the comments of this page http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/267197 ) showing how to encode a username/password into a request header. It looks like this :

import base64
base64string = base64.encodestring('%s:%s' % (data['name'], data['pass']))[:-1]
req.add_header("Authorization", "Basic %s" % base64string)

Where req is our request object like in the first example.

Let's wrap all this up with an example that shows accessing a page, doing the authentication and saving the realm. I use a regular expression to pull the scheme and realm out of the authentication response header. I use urlparse to get the server part of the url. If we store the username/password (or the whole request header line) then we can re-use that information automatically if we come across another page in that realm. When the code has run the contents of the page we've fetched is saved as a string in the variable 'thepage'. If you are writing an http client of any sort that has to deal with basic authentication then this example should have everything you need to know - but see the comment below about cookies.

Some websites may also use cookies with authentication. Luckily there is a library that will allow you to have automatic cookie management without thinking about it. This is ClientCookie (http://wwwsearch.sourceforge.net/ClientCookie/ ). In python 2.4 it becomes part of the python standard library as clientcookie. See my cookbook example of how to use it (http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/302930 ). I've done a bigger example that displays a lot of http information including cookies, headers, environment variables (the CGI environment) and all this authentication stuff. You can find it at http://www.voidspace.org.uk/atlantibots/recipebook.html#http - it's a CGI and there's an online version to try. It shows lot's of the information about http state, the CGI environment etc..

--

In actual fact the 'proper' way to do BASIC authentication with Python is to install an authentication handler as an 'opener' (along with a password manager) in urllib2. It doesn't show as clearly what is happening and is less suitable for my needs (within a CGI), where a pickled dictionary is more usefl. See at the bottom of this example for an alternative example using an auth handler - this was sent to me by Jaime Wyant (and amended by me)....
"""

import urllib2, sys, re, base64
from urlparse import urlparse
theurl = 'http://www.someserver.com/somepath/somepage.html'            # if you want to run this example you'll need to supply a protected page with your username and password
username = 'johnny'
password = 'XXXXXX'            # a very bad password
theurl = 'http://freeway.resheteva.lan/~alex/googlepages'            # if you want to run this example you'll need to supply a protected page with your username and password
username = 'alex'
password = 'rama0ET4'

req = urllib2.Request(theurl)
try:
    handle = urllib2.urlopen(req)
except IOError, e:                  # here we are assuming we fail
    pass
else:                               # If we don't fail then the page isn't protected
    print "This page isn't protected by authentication."
    sys.exit(1)
    
if not hasattr(e, 'code') or e.code != 401:                 # we got an error - but not a 401 error
    print "This page isn't protected by authentication."
    print 'But we failed for another reason.'
    sys.exit(1)

authline = e.headers.get('www-authenticate', '')                # this gets the www-authenticat line from the headers - which has the authentication scheme and realm in it
if not authline:
    print 'A 401 error without an authentication response header - very weird.'
    sys.exit(1)
    
authobj = re.compile(r'''(?:\s*www-authenticate\s*:)?\s*(\w*)\s+realm=['"](\w+)['"]''', re.IGNORECASE)          # this regular expression is used to extract scheme and realm
authobj = re.compile(r'\s*(\w*)\s+realm=[\'"](\w+)[\'"]', re.IGNORECASE)          # this regular expression is used to extract scheme and realm
matchobj = authobj.match(authline)
if not matchobj:                                        # if the authline isn't matched by the regular expression then something is wrong
    print 'The authentication line is badly formed.', authline
#    sys.exit(1)
#scheme = matchobj.group(1) 
#realm = matchobj.group(2)
#if scheme.lower() != 'basic':
#    print 'This example only works with BASIC authentication.'
#    sys.exit(1)

base64string = base64.encodestring('%s:%s' % (username, password))[:-1]
authheader =  "Basic %s" % base64string
req.add_header("Authorization", authheader)
try:
    handle = urllib2.urlopen(req)
except IOError, e:                  # here we shouldn't fail if the username/password is right
    print "It looks like the username or password is wrong."
    sys.exit(1)
thepage = handle.read()
    
server = urlparse(theurl)[1].lower()            # server names are case insensitive, so we will convert to lower case
test = server.find(':')
if test != -1: server = server[:test]           # remove the :port information if present, we're working on the principle that realm names per server are likely to be unique...

realm=None
passdict = {(server, realm) : authheader }      # now if we get another 401 we can test for an entry in passdict before having to ask the user for a username/password

print 'Done successfully - information now stored in passdict.'
print 'The webpage is stored in thepage.'

"""
The proper way of actually doing all this is to use an HTTPBasicAuthHandler along with an HTTPPasswordMgr.
The python documentation on this is actually pretty minimal - so below is an example showing how to do this.
The main problem with HTTPPasswordMgr is that you must *already* know the realm - so we're going to use the HTTPPasswordMgrWithDefaultRealm instead !!

theurl = 'http://www.someserver.com/highestlevelprotectedpath/somepage.htm'
username = 'johnny'
password = 'XXXXXX'         # a great password

passman = urllib2.HTTPPasswordMgrWithDefaultRealm()      # this creates a password manager
passman.add_password(None, theurl, username, password)      # because we have put None at the start it will always use this username/password combination

authhandler = urllib2.HTTPBasicAuthHandler(passman)                 # create the AuthHandler

opener = urllib2.build_opener(authhandler)                                  # build an 'opener' using the handler we've created
# you can use the opener directly to open URLs
# *or* you can install it as the default opener so that all calls to urllib2.urlopen use this opener
urllib2.install_opener(opener)

# All calls to urllib2.urlopen will now use our handler


ISSUES


CHANGELOG
16-09-04        Version 1.0.0
I think it's ok.
"""



# -*- coding: utf-8 -*-

"""enables to import modules through the web, by adding urls to the python path.
"""
#authors: Jure Vrscaj <jure@codeshift.net> (copyright (c) 2006-10),
#         송영진(MyEvan) <myevan_net@naver.com> (copyright (c) 2010), 
#         Alex Bodnaru <alexbodn@012.net.il> (copyright (c) 2010)

# This is free software; you may use, copy, modify and/or distribute this work
# under the terms of the MIT License.


import sys
from urlimport import UrlFinder, config, reset, DefaultErrorHandler

__all__ = ('config', 'reset', 'DefaultErrorHandler')

# register The Hook
sys.path_hooks = \
    [x for x in sys.path_hooks if x.__name__ != 'UrlFinder'] + [UrlFinder]






import doctest
import test_urlimport
doctest.testmod(test_urlimport, optionflags=doctest.NORMALIZE_WHITESPACE |
                                   doctest.ELLIPSIS |
                                   doctest.REPORT_ONLY_FIRST_FAILURE
                                   )



#coding: utf-8

from setuptools import setup, find_packages
import sys, os

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

def readlines(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).readlines()

pyversion = sys.version_info[0] * 10 + sys.version_info[1]

setup(name='urlimport',
      version=[l for l in readlines('changelog') \
               if l.strip() and not l.lstrip().startswith('#')][0].split()[0][1:],
      description="Allow importing files through the web, by adding urls to pythopath",
      long_description=read('README'),
      platforms='ALL',
      keywords='python import remote modules http https ftp perforce',
      author='Jure Vrscaj, 송영진(MyEvan), Alex Bodnaru',
      author_email='jure@codeshift.net, myevan_net@naver.com, alexbodn@012.net.il',
      maintainer='alex bodnaru',
      maintainer_email='alexbodn@012.net.il',
      url='http://packages.python.org/urlimport',
      license='MIT',
      packages=find_packages(exclude=['ez_setup', 'server_side', 'tests']),
      include_package_data=True,
      zip_safe=True,
      install_requires=[
          # -*- Extra requirements: -*-
          "pysqlite"
      ] if pyversion < 25 else [],
      entry_points="""
      # -*- Entry points: -*-
      """,
      # not yet there ;)
      #test_suite = 'nose.collector',
      classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 2", # as of now
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Topic :: Communications",
        "Topic :: Internet",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: System :: Distributed Computing",
        "License :: OSI Approved :: MIT License",
      ],
#      dist_files=['changelog'],
)




# -*- coding: utf-8 -*-

"""enables to import modules from perforce, by adding p4:// urls to the python 
   path.
"""
#authors: Jure Vrscaj <jure@codeshift.net> (copyright (c) 2006-10),
#         송영진(MyEvan) <myevan_net@naver.com> (copyright (c) 2010), 
#         Alex Bodnaru <alexbodn@012.net.il> (copyright (c) 2010)

# This is free software; you may use, copy, modify and/or distribute this work
# under the terms of the MIT License.


from urlimport import UrlFinder, debug
import sys, re
from P4 import P4
from datetime import datetime

class PerforceFinder(UrlFinder):

    re_url_ok = re.compile(r'^p4://')

    connection_cache = dict()

    def get_file_data(self, url, cached_last_modified, cached_etag):
        """Download the file data from given url.
        """
        p4conn = self.connection_cache.get(self.path, None)
        if p4conn is None:
            p4conn = P4()
            p4conn.connect()
            self.connection_cache[self.path] = p4conn
            debug("p4 connect", lvl=1)

        fstats = p4conn.run_fstat(url[3:])
        for fstat in fstats:
            last_modified = datetime.fromtimestamp(int(fstat["headModTime"]))
            keep_cached = last_modified <= self.parse_date(cached_last_modified) \
                if cached_last_modified else False
            last_modified = last_modified.strftime(self.date_format)
            if not keep_cached:
                try:
                    fileData = open(fstat["clientFile"]).read()
                except IOError:
                    p4conn.run_sync(url[3:])
                    fileData = open(fstat["clientFile"]).read()
            else:
                fileData = None

            return fileData, \
                   None,\
                   last_modified,\
                   keep_cached,\
                   None, \
                   url

    def split_url(self, url):
        return ("",) * 8


# register The Hook
sys.path_hooks = \
    [x for x in sys.path_hooks if x.__name__ != 'PerforceFinder'] + [PerforceFinder]

